from typing import Tuple, List
from bencher.bench_vars import ResultVar, ResultVec, ResultList, ResultSeries
from collections import namedtuple
import param


def make_namedtuple(class_name: str, **fields) -> namedtuple:
    """Convenience method for making a named tuple

    Args:
        class_name (str): name of the named tuple

    Returns:
        namedtuple: a named tuple with the fields as values
    """
    return namedtuple(class_name, fields)(*fields.values())


def update_params_from_kwargs(self, **kwargs) -> None:
    """Given a dictionary of kwargs, set the parameters of the passed class 'self' to the values in the dictionary."""
    used_params = {}
    for key in self.param.params().keys():
        if key in kwargs:
            if key != "name":
                used_params[key] = kwargs[key]

    self.param.update(**used_params)


def get_input_and_results(self, include_name: bool = False) -> Tuple[dict, dict]:
    """Get dictionaries of input parameters and result parameters

    Args:
        self: A parametrised class
        include_name (bool): Include the name parameter that all parametrised classes have. Default False

    Returns:
        Tuple[dict, dict]: a tuple containing the inputs and result parameters as dictionaries
    """
    inputs = {}
    results = {}
    for k, v in self.param.params().items():
        if isinstance(v, (ResultList, ResultSeries, ResultVar, ResultVec)):
            results[k] = v
        else:
            inputs[k] = v

    if not include_name:
        inputs.pop("name")
    return make_namedtuple("inputresult", inputs=inputs, results=results)


def get_inputs_only(self) -> List[param.Parameter]:
    """Return a list of input parameters

    Returns:
        List[param.Parameter]: A list of input parameters
    """
    return list(get_input_and_results(self).inputs.values())


def get_results_only(self) -> List[param.Parameter]:
    """Return a list of input parameters

    Returns:
        List[param.Parameter]: A list of result parameters
    """
    return list(get_input_and_results(self).results.values())


def get_results_values_as_dict(self) -> dict:
    """Get a dictionary of result variables with the name and the current value"""
    values = self.param.values()
    return {key: values[key] for key in get_input_and_results(self).results}



# from bencher import Pa#
import param
import xarray


class PlotBase:
    def title(self, x: param.Parameter, y: param.Parameter, z: param.Parameter = None) -> str:
        if z is None:
            return f"{x.name} vs {y.name}"
        return f"{z.name} vs ({x.name} vs {y.name})"

    def calculate_stats(self, da, over_dim="repeat"):
        mean = da.mean(over_dim)
        std = da.std(over_dim)

        low = mean + std
        high = mean - std
        mean = mean.rename("mean")
        low = low.rename("std_low")
        high = high.rename("std_high")

        return xarray.merge([mean, low, high]), std



import random

import bencher as bch


class GaussianDist(bch.ParametrizedSweep):
    """A class to represent a gaussian distribution."""

    mean = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], doc="mean of the gaussian distribution", samples=3
    )
    sigma = bch.FloatSweep(
        default=1, bounds=[0, 1.0], doc="standard deviation of gaussian distribution", samples=4
    )


class Example2DGaussianResult(bch.ParametrizedSweep):
    """A class to represent the properties of a volume sample."""

    gauss_x = bch.ResultVar("m", doc="x value of the 2D gaussian")
    gauss_y = bch.ResultVar("m", doc="y value of the 2D gaussian")

    point2D = bch.ResultVec(2, "m", doc="2D vector of the point")


def bench_fn(dist: GaussianDist) -> Example2DGaussianResult:
    """This function samples a point from a gaussian distribution.

    Args:
        dist (GaussianDist): Sample point

    Returns:
        Example2DGaussianResult: Value at that point
    """
    output = Example2DGaussianResult()

    output.gauss_x = random.gauss(dist.mean, dist.sigma)
    output.gauss_y = random.gauss(dist.mean, dist.sigma)
    output.point2D = [output.gauss_x, output.gauss_y]

    return output


def example_floats2D_scatter(run_cfg: bch.BenchRunCfg) -> bch.Bench:
    """Example of how to perform a 3D floating point parameter sweep

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """
    bench = bch.Bench(
        "Bencher_Example_Floats_Scatter", bench_fn, GaussianDist, plot_lib=bch.PlotLibrary.default()
    )

    bench.plot_sweep(
        result_vars=[
            Example2DGaussianResult.param.point2D,
            Example2DGaussianResult.param.gauss_x,
            Example2DGaussianResult.param.gauss_y,
        ],
        title="Float 2D Scatter Example",
        run_cfg=run_cfg,
    )

    bench.plot_sweep(
        input_vars=[GaussianDist.param.mean],
        result_vars=[
            Example2DGaussianResult.param.point2D,
            Example2DGaussianResult.param.gauss_x,
            Example2DGaussianResult.param.gauss_y,
        ],
        title="Float 2D Scatter With Changing Mean",
        run_cfg=run_cfg,
    )

    bench.plot_sweep(
        input_vars=[GaussianDist.param.sigma],
        result_vars=[
            Example2DGaussianResult.param.point2D,
            Example2DGaussianResult.param.gauss_x,
            Example2DGaussianResult.param.gauss_y,
        ],
        title="Float 2D Scatter With Changing Sigma",
        run_cfg=run_cfg,
    )

    # future work
    # bench.plot_sweep(
    #     input_vars=[GaussianDist.param.mean, GaussianDist.param.sigma],
    #     result_vars=[
    #         GaussianResult.param.point2D,
    #         GaussianResult.param.gauss_x,
    #         GaussianResult.param.gauss_y,
    #     ],
    #     title="Float 2D Scatter With Changing Sigma",
    #     run_cfg=run_cfg,
    # )

    return bench


if __name__ == "__main__":
    ex_run_cfg = bch.BenchRunCfg()
    ex_run_cfg.repeats = 50
    ex_run_cfg.over_time = True
    # ex_run_cfg.clear_history = True
    example_floats2D_scatter(ex_run_cfg).plot()



import bencher as bch
import unittest


class ExampleClass(bch.ParametrizedSweep):
    iv1 = bch.FloatSweep()
    rv1 = bch.ResultVar()


class TestBencherUtils(unittest.TestCase):
    def test_get_inputs(self) -> None:
        ex_instance = ExampleClass()
        inputs = bch.get_inputs_only(ex_instance)

        print(inputs)

        self.assertEqual(len(inputs), 1)
        self.assertEqual(inputs[0].name, "iv1")

    def test_get_results(self) -> None:
        ex_instance = ExampleClass()
        results = bch.get_results_only(ex_instance)

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].name, "rv1")

    def test_get_inputs_and_results(self) -> None:
        ex_instance = ExampleClass()
        inputs, results = bch.get_input_and_results(ex_instance)

        self.assertEqual(len(inputs), 1)
        self.assertEqual(len(results), 1)
        self.assertEqual(inputs["iv1"].name, "iv1")
        self.assertEqual(results["rv1"].name, "rv1")

    def test_get_results_values_as_dict(self) -> None:
        ex_instance = ExampleClass()

        ex_instance.rv1 = 3

        res = bch.get_results_values_as_dict(ex_instance)

        self.assertEqual(res["rv1"], 3)



import errno
import logging
import os
from textwrap import wrap

import holoviews as hv
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import panel as pn
import plotly.express as px
import seaborn as sns
from holoviews import opts

from bencher.bench_cfg import BenchCfg, PltCfgBase
from bencher.bench_vars import ParametrizedSweep, ResultList, ResultVar, ResultVec
import plotly.graph_objs as go


hv.extension("bokeh", "plotly")
# hv.extension("plotly", "bokeh")

# hv.extension("plotly")


def wrap_long_time_labels(bench_cfg: BenchCfg) -> BenchCfg:
    """Takes a benchCfg and wraps any index labels that are too long to be plotted easily

    Args:
        bench_cfg (BenchCfg):

    Returns:
        BenchCfg: updated config with wrapped labels
    """
    if bench_cfg.over_time:
        if bench_cfg.ds.coords["over_time"].dtype == np.datetime64:
            # plotly catastrophically fails to plot anything with the default long string representation of time, so convert to a shorter time representation
            bench_cfg.ds.coords["over_time"] = [
                pd.to_datetime(t).strftime("%d-%m-%y %H-%M-%S")
                for t in bench_cfg.ds.coords.coords["over_time"].values
            ]
            # wrap very long time event labels because otherwise the graphs are unreadable
        if bench_cfg.time_event is not None:
            bench_cfg.ds.coords["over_time"] = [
                "\n".join(wrap(t, 20)) for t in bench_cfg.ds.coords["over_time"].values
            ]
    return bench_cfg


def plot_sns(bench_cfg: BenchCfg, rv: ParametrizedSweep, sns_cfg: PltCfgBase) -> pn.pane:
    """Plot with seaborn

    Args:
        bench_cfg (BenchCfg): bench config
        rv (ParametrizedSweep): the result variable to plot
        sns_cfg (PltCfgBase): the plot configuration

    Returns:
        pn.pane: A seaborn plot as a panel pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)

    if bench_cfg.use_holoview:
        opts.defaults(
            opts.Curve(
                colorbar=True,
                colorbar_opts={"title": rv.name},
                width=600,
                height=600,
                xlabel=sns_cfg.xlabel,
                ylabel=sns_cfg.ylabel,
                title=sns_cfg.title,
            )
        )
        ds = hv.Dataset(bench_cfg.ds[rv.name].mean("repeat"))

        if bench_cfg.over_time:
            agg = ds.aggregate("over_time", np.mean, spreadfn=np.std)
            plot = hv.Spread(agg) * hv.Curve(agg)
        else:
            plot = ds.to(hv.Bars) * ds.to(hv.Scatter).opts(jitter=50)
        return plot
    plt.rcParams.update({"figure.max_open_warning": 0})

    # if type(rv) == ResultVec:
    # return plot_scatter_sns(bench_cfg, rv)
    # else:

    if type(rv) == ResultVec:
        if rv.size == 2:
            plt.figure(figsize=(4, 4))
            fg = plot_scatter2D_sns(bench_cfg, rv)
        elif rv.size == 3:
            return plot_scatter3D_px(bench_cfg, rv)
        else:
            return pn.pane.Markdown("Scatter plots of >3D result vectors not supported yet")
    elif type(rv) == ResultVar:
        df = bench_cfg.ds[rv.name].to_dataframe().reset_index()

        try:
            fg = sns_cfg.plot_callback(data=df, **sns_cfg.as_sns_args())
        except Exception as e:
            return pn.pane.Markdown(
                f"Was not able to plot becuase of exception:{e} \n this is likely due to too many NAN values"
            )

        # TODO try to set this during the initial plot rather than after
        for ax in fg.axes.flatten():
            for tick in ax.get_xticklabels():
                tick.set_rotation(45)

        fg.set_xlabels(label=sns_cfg.xlabel, clear_inner=True)
        fg.set_ylabels(label=sns_cfg.ylabel, clear_inner=True)
    elif type(rv) == ResultList:
        plt.figure(figsize=(4, 4))

        in_var = bench_cfg.input_vars[0].name

        # get the data to plot
        dataset_key = (bench_cfg.hash_value, rv.name)
        df = bench_cfg.ds_dynamic[dataset_key].to_dataframe().reset_index()

        # plot
        fg = sns.lineplot(df, x=rv.dim_name, y=rv.name, hue=in_var)

        # titles and labels and formatting
        fg.set_xlabel(f"{rv.dim_name} [{rv.dim_units}]")
        fg.set_ylabel(f"{rv.name} [{rv.units}]")
        fg.set_title(f"{in_var} vs ({rv.name} vs {rv.dim_name})")

        plt.tight_layout()

        return pn.panel(plt.gcf())

    fg.fig.suptitle(sns_cfg.title)
    plt.tight_layout()

    if bench_cfg.save_fig:
        save_fig(bench_cfg, sns_cfg)
    return pn.panel(plt.gcf())


def plot_scatter2D_sns(bench_cfg: BenchCfg, rv: ParametrizedSweep) -> pn.pane.Plotly:
    """Given a benchCfg generate a 2D scatter plot from seaborn

    Args:
        bench_cfg (BenchCfg): description of benchmark
        rv (ParametrizedSweep): result variable to plot

    Returns:
        pn.pane.Plotly: A 3d volume plot as a holoview in a pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)
    ds = bench_cfg.ds.drop_vars("repeat")

    df = ds.to_dataframe().reset_index()

    names = rv.index_names()

    if bench_cfg.input_vars:
        h = sns.jointplot(df, x=names[0], y=names[1], hue=bench_cfg.input_vars[0].name)
    elif bench_cfg.over_time:
        h = sns.jointplot(df, x=names[0], y=names[1], hue=bench_cfg.iv_time[0].name)

    else:
        h = sns.jointplot(df, x=names[0], y=names[1])

    h.set_axis_labels(f"{names[0]} [{rv.units}]", f"{names[1]} [{rv.units}]")
    return h


def plot_scatter2D_hv(bench_cfg: BenchCfg, rv: ParametrizedSweep) -> pn.pane.Plotly:
    """Given a benchCfg generate a 2D scatter plot

    Args:
        bench_cfg (BenchCfg): description of benchmark
        rv (ParametrizedSweep): result variable to plot

    Returns:
        pn.pane.Plotly: A 3d volume plot as a holoview in a pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)
    bench_cfg.ds.drop_vars("repeat")

    df = bench_cfg.get_dataframe()

    names = rv.index_names()

    return px.scatter(df, x=names[0], y=names[1], marginal_x="histogram", marginal_y="histogram")


def plot_scatter3D_px(bench_cfg: BenchCfg, rv: ParametrizedSweep) -> pn.pane.Plotly:
    """Given a benchCfg generate a 3D scatter plot with plotly express

    Args:
        bench_cfg (BenchCfg): description of benchmark
        rv (ParametrizedSweep): result variable to plot

    Returns:
        pn.pane.Plotly: A 3d scatter plot as a holoview in a pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)

    df = bench_cfg.get_dataframe()

    names = rv.index_names()  # get the column names of the vector result

    if bench_cfg.input_vars:
        color = bench_cfg.input_vars[0].name
    else:
        color = None

    if bench_cfg.over_time:
        if len(names) < 3:  # only a 2d vector result so use the time axis as the third point
            names.insert(0, bench_cfg.iv_time[0].name)

    return px.scatter_3d(
        df, x=names[0], y=names[1], z=names[2], color=color, symbol=color, size_max=2
    )


def save_fig(
    bench_cfg: BenchCfg,
    sns_cfg: PltCfgBase,
):
    """Save a seaborn figure to disk based on bench config data

    Args:
        bench_cfg (BenchCfg): benchmark config
        sns_cfg (PltCfgBase): plotting config
    """
    sns_cfg.data = None  # remove data before getting __repr__ for filename
    figname = f"{bench_cfg.as_filename()},{sns_cfg.as_filename()}"
    figpath = os.path.join("autofig", bench_cfg.bench_name, f"{figname}.png")
    logging.info(f"saving:{figpath}")
    dir_name = os.path.dirname(figpath)
    if not os.path.exists(dir_name) and dir_name != "":
        try:
            os.makedirs(dir_name)
        except OSError as exc:  # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise

    if os.path.exists(figpath) and bench_cfg.raise_duplicate_exception:
        raise FileExistsError(
            f"This figname {figname} already exists, please define a unique benchmark name or don't run the same benchmark twice"
        )
    plt.savefig(figpath)


def plot_surface_plotly(
    bench_cfg: BenchCfg, rv: ParametrizedSweep, xr_cfg: PltCfgBase
) -> pn.pane.Plotly:
    """Given a benchCfg generate a 2D surface plot
    Args:
        bench_cfg (BenchCfg): description of benchmark
        rv (ParametrizedSweep): result variable to plot
        xr_cfg (PltCfgBase): config of x,y variables
    Returns:
        pn.pane.Plotly: A 2d surface plot as a holoview in a pane
    """

    if type(rv) == ResultVec:
        return plot_scatter3D_px(bench_cfg, rv)

    bench_cfg = wrap_long_time_labels(bench_cfg)

    da = bench_cfg.ds[rv.name].transpose()

    mean = da.mean("repeat")

    x = da.coords[xr_cfg.x]
    y = da.coords[xr_cfg.y]

    opacity = 0.3

    surfaces = [go.Surface(x=x, y=y, z=mean)]

    if bench_cfg.repeats > 1:
        std_dev = da.std("repeat")
        surfaces.append(go.Surface(x=x, y=y, z=mean + std_dev, showscale=False, opacity=opacity))
        surfaces.append(go.Surface(x=x, y=y, z=mean - std_dev, showscale=False, opacity=opacity))

    eye_dis = 1.7
    layout = go.Layout(
        title=xr_cfg.title,
        width=700,
        height=700,
        scene=dict(
            xaxis_title=xr_cfg.xlabel,
            yaxis_title=xr_cfg.ylabel,
            zaxis_title=xr_cfg.zlabel,
            camera={"eye": {"x": eye_dis, "y": eye_dis, "z": eye_dis}},
        ),
    )

    fig = {"data": surfaces, "layout": layout}

    return pn.pane.Plotly(fig)


def plot_surface_holo(
    bench_cfg: BenchCfg, rv: ParametrizedSweep, xr_cfg: PltCfgBase
) -> pn.pane.Plotly:
    """Given a benchCfg generate a 2D surface plot
    Args:
        bench_cfg (BenchCfg): description of benchmark
        rv (ParametrizedSweep): result variable to plot
        xr_cfg (PltCfgBase): config of x,y variables
    Returns:
        pn.pane.holoview: A 2d surface plot as a holoview in a pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)

    alpha = 0.3

    da = bench_cfg.ds[rv.name]

    # hv.extension("plotly")
    # hv.extension("bokeh", "plotly")

    mean = da.mean("repeat")

    # TODO a warning suggests setting this parameter, but it does not seem to help as expected, leaving here to fix in the future
    # hv.config.image_rtol = 1.0

    ds = hv.Dataset(mean)
    surface = ds.to(hv.Surface)

    if bench_cfg.repeats > 1:
        std_dev = da.std("repeat")
        surface *= (
            hv.Dataset(mean + std_dev)
            .to(hv.Surface)
            .opts(alpha=alpha, colorbar=False, backend="plotly")
        )
        surface *= (
            hv.Dataset(mean - std_dev)
            .to(hv.Surface)
            .opts(alpha=alpha, colorbar=False, backend="plotly")
        )
    return pn.Column(
        # using render disabled the holoviews sliders :(
        hv.render(
            surface.opts(
                width=800, height=800, zlabel=xr_cfg.zlabel, title=xr_cfg.title, backend="plotly"
            ),
            backend="plotly",
        )
    )



from __future__ import annotations

from bencher.plotting.plot_collection import PlotCollection
from bencher.plotting.plot_types import PlotTypes
from bencher.plotting.plots.catplot import Catplot
from bencher.plotting.plots.heatmap import Heatmap
from bencher.plotting.plots.lineplot import Lineplot
from bencher.plotting.plots.scatterplot import Scatter
from bencher.plotting.plots.tables import Tables
from bencher.plotting.plots.volume import VolumePlot
from bencher.plotting.plots.surface import SurfacePlot

from bencher.plotting.plots.hv_interactive import HvInteractive


class PlotLibrary:
    """This class is a collection of static factory methods that define sets of plots.  It provides a sensible default PlotCollection out of the box but also enables the creation of custom PlotCollections"""

    @staticmethod
    def setup_sources() -> PlotCollection:
        """Shared method that collects all the possible plot types.  This is used by all the other methods that select different subsections of the plots

        Returns:
            PlotCollection: A PlotCollection with sources of plots defined, but none selected
        """
        plt_col = PlotCollection()
        plt_col.add_plotter_source(Catplot())
        plt_col.add_plotter_source(Tables())
        plt_col.add_plotter_source(Heatmap())
        plt_col.add_plotter_source(Lineplot())
        plt_col.add_plotter_source(Scatter())
        plt_col.add_plotter_source(HvInteractive())
        plt_col.add_plotter_source(VolumePlot())
        plt_col.add_plotter_source(SurfacePlot())

        return plt_col

    @staticmethod
    def default() -> PlotCollection:
        """Default PlotCollection.  This set of plots should provide a reasonable set of plots that cover the most common data plotting use cases

        Returns:
            PlotCollection: sensible default for PlotCollection
        """
        plt_col = PlotLibrary.setup_sources()
        # plt_col.add(PlotTypes.scatter_hv)

        plt_col.add(PlotTypes.swarmplot)
        plt_col.add(PlotTypes.heatmap_2D)  # 2d image of a surface
        plt_col.add(PlotTypes.lineplot)
        plt_col.add(PlotTypes.lineplot_hv)
        plt_col.add(PlotTypes.lineplot_hv_overlay)
        plt_col.add(PlotTypes.lineplot_hv_layout)
        plt_col.add(PlotTypes.scatter_hv)
        plt_col.add(PlotTypes.bar_hv)
        plt_col.add(PlotTypes.volume_plotly)
        plt_col.add(PlotTypes.cone_plotly)
        # plt_col.add(PlotTypes.lineplot_hv_subplot)
        plt_col.add(PlotTypes.scatter2D_sns)
        plt_col.add(PlotTypes.surface_hv)
        # plt_col.add(PlotTypes.hv_interactive)

        return plt_col

    @staticmethod
    def tables() -> PlotCollection:
        """Display only table summaries of the data

        Returns:
            PlotCollection: Only table plots
        """
        plt_col = PlotLibrary.setup_sources()
        plt_col.add(PlotTypes.dataframe_multi_index)
        plt_col.add(PlotTypes.dataframe_mean)
        return plt_col

    @staticmethod
    def all() -> PlotCollection:
        """Display the data in every possible way. All plots are included

        Returns:
            PlotCollection: All possible plots
        """
        plt_col = PlotLibrary.setup_sources()
        for pt in PlotTypes:  # iterate through all plot enums
            plt_col.add(pt)
        return plt_col

    @staticmethod
    def none() -> PlotCollection:
        """Set up a plot collection with no active plots

        Returns:
            PlotCollection: No plots
        """
        return PlotLibrary.setup_sources()



import logging
from datetime import datetime
from itertools import product
from typing import Callable, List

import numpy as np
import panel as pn
import param
import xarray as xr
from diskcache import Cache
from sortedcontainers import SortedDict

from bencher.bench_cfg import BenchCfg, BenchRunCfg, DimsCfg
from bencher.bench_plot_server import BenchPlotServer
from bencher.bench_vars import (
    IntSweep,
    ParametrizedSweep,
    ResultList,
    ResultVar,
    ResultVec,
    TimeEvent,
    TimeSnapshot,
    hash_sha1,
)
from bencher.plotting.plot_collection import PlotCollection
from bencher.plt_cfg import BenchPlotter
from bencher.plotting.plot_library import PlotLibrary  # noqa pylint: disable=unused-import

# Customize the formatter
formatter = logging.Formatter("%(levelname)s: %(message)s")
logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")

for handler in logging.root.handlers:
    handler.setFormatter(formatter)


def set_xarray_multidim(data_array: xr.DataArray, index_tuple, value: float) -> xr.DataArray:
    """This is terrible, I need to do this in a better way, but [] does not like *args syntax and the () version of the set function doesn't either"""
    match len(index_tuple):
        case 1:
            data_array[index_tuple[0]] = value
        case 2:
            data_array[index_tuple[0], index_tuple[1]] = value
        case 3:
            data_array[index_tuple[0], index_tuple[1], index_tuple[2]] = value
        case 4:
            data_array[index_tuple[0], index_tuple[1], index_tuple[2], index_tuple[3]] = value
        case 5:
            data_array[
                index_tuple[0], index_tuple[1], index_tuple[2], index_tuple[3], index_tuple[4]
            ] = value
        case 6:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
            ] = value
        case 7:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
                index_tuple[6],
            ] = value
        case 8:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
                index_tuple[6],
                index_tuple[7],
            ] = value
        case 9:
            data_array[
                index_tuple[0],
                index_tuple[1],
                index_tuple[2],
                index_tuple[3],
                index_tuple[4],
                index_tuple[5],
                index_tuple[6],
                index_tuple[7],
                index_tuple[8],
            ] = value
    return data_array


class Bench(BenchPlotServer):
    def __init__(
        self,
        bench_name: str = None,
        worker: Callable = None,
        worker_input_cfg: ParametrizedSweep = None,
        plot_lib: PlotCollection = PlotLibrary.default(),
    ) -> None:
        """Create a new Bench object from a function and a class defining the inputs to the function

        Args:
            bench_name (str): The name of the benchmark and output folder for the figures
            worker (Callable): A function that accepts a class of type (worker_input_config)
            worker_input_config (ParametrizedSweep): A class defining the parameters of the function.
            plot_lib: (PlotCollection):  A dictionary of plot names:method pairs that are selected for plotting based on the type of data they can plot.
        """
        self.bench_name = bench_name
        self.worker = None
        self.worker_input_cfg = None
        self.set_worker(worker, worker_input_cfg)

        self.pane = pn.Tabs(tabs_location="left", name=self.bench_name)
        # The number of times the wrapped worker was called
        self.worker_wrapper_call_count = 0
        self.worker_fn_call_count = 0  # The number of times the raw worker was called
        # The number of times the cache was used instead of the raw worker
        self.worker_cache_call_count = 0
        self.bench_cfg_hashes = []  # a list of hashes that point to benchmark results
        self.last_run_cfg = None  # cached run_cfg used to pass to the plotting function
        self.sample_cache = None  # store the results of each benchmark function call in a cache
        self.ds_dynamic = {}  # A dictionary to store unstructured vector datasets
        self.plot_lib = plot_lib

    def set_worker(self, worker: Callable, worker_input_cfg: ParametrizedSweep = None) -> None:
        """Set the benchmark worker function and optionally the type the worker expects

        Args:
            worker (Callable): The benchmark worker function
            worker_input_cfg (ParametrizedSweep, optional): The input type the worker expects. Defaults to None.
        """
        self.worker = worker
        self.worker_input_cfg = worker_input_cfg

    def plot_sweep(
        self,
        title: str,
        input_vars: List[ParametrizedSweep] = None,
        result_vars: List[ParametrizedSweep] = None,
        const_vars: List[ParametrizedSweep] = None,
        time_src: datetime = None,
        description: str = None,
        post_description: str = None,
        pass_repeat: bool = False,
        tag: str = "",
        run_cfg: BenchRunCfg = None,
        plot_lib=None,
    ) -> BenchCfg:
        """The all in 1 function benchmarker and results plotter.

        Args:
            input_vars (List[ParametrizedSweep], optional): _description_. Defaults to None.
            result_vars (List[ParametrizedSweep], optional): _description_. Defaults to None.
            const_vars (List[ParametrizedSweep], optional): A list of variables to keep constant with a specified value. Defaults to None.
            title (str, optional): The title of the benchmark. Defaults to None.
            description (str, optional): A description of the benchmark. Defaults to None.
            post_description (str, optional): A description that comes after the benchmark plots. Defaults to None.
            time_src (datetime, optional): Set a time that the result was generated. Defaults to datetime.now().
            pass_repeat (bool,optional) By default do not pass the kwarg 'repeat' to the benchmark function.  Set to true if
            you want the benchmark function to be passed the repeat number
            tag (str,optional): Use tags to group different benchmarks together.
            run_cfg: (BenchRunCfg, optional): A config for storing how the benchmarks and run and plotted
            plot_lib: (PlotCollection):  A dictionary of plot names:method pairs that are selected for plotting based on the type of data they can plot.

        Raises:
            ValueError: If a result variable is not set

        Returns:
            BenchCfg: A class with all the data used to generate the results and the results
        """

        if input_vars is None:
            input_vars = []
        if result_vars is None:
            result_vars = []
        if const_vars is None:
            const_vars = []

        for i in input_vars:
            self.check_var_is_a_param(i, "input")
        for i in result_vars:
            self.check_var_is_a_param(i, "result")
        for i in const_vars:
            # consts come as tuple pairs
            self.check_var_is_a_param(i[0], "const")

        if post_description is None:
            post_description = (
                "## Description\nPlease set post_description to explain these results"
            )
        if run_cfg is None:
            run_cfg = BenchRunCfg()
        elif run_cfg.only_plot:
            run_cfg.use_cache = True
        self.last_run_cfg = run_cfg

        bench_cfg = BenchCfg(
            input_vars=input_vars,
            result_vars=result_vars,
            const_vars=const_vars,
            bench_name=self.bench_name,
            description=description,
            post_description=post_description,
            title=title,
            pass_repeat=pass_repeat,
            tag=tag,
        )

        bench_cfg.param.update(run_cfg.param.values())
        bench_cfg.plot_lib = plot_lib if plot_lib is not None else self.plot_lib

        print("plot_lib", bench_cfg.plot_lib)

        bench_cfg_hash = bench_cfg.hash_persistent(True)
        bench_cfg.hash_value = bench_cfg_hash

        # does not include repeats in hash as sample_hash already includes repeat as part of the per sample hash
        bench_cfg_sample_hash = bench_cfg.hash_persistent(False)

        if bench_cfg.use_sample_cache:
            self.sample_cache = Cache("cachedir/sample_cache", tag_index=True)
            if bench_cfg.clear_sample_cache:
                self.clear_tag_from_cache(bench_cfg.tag)

        calculate_results = True
        with Cache("cachedir/benchmark_inputs") as c:
            if run_cfg.clear_cache:
                c.delete(bench_cfg_hash)
                logging.info("cleared cache")
            elif run_cfg.use_cache:
                logging.info(
                    f"checking for previously calculated results with key: {bench_cfg_hash}"
                )
                if bench_cfg_hash in c:
                    logging.info(f"loading cached results from key: {bench_cfg_hash}")
                    bench_cfg = c[bench_cfg_hash]
                    # if not over_time:  # if over time we always want to calculate results
                    calculate_results = False
                else:
                    logging.info("did not detect results in cache")
                    if run_cfg.only_plot:
                        raise FileNotFoundError("Was not able to load the results to plot!")

        if calculate_results:
            if run_cfg.time_event is not None:
                time_src = run_cfg.time_event
            bench_cfg = self.calculate_benchmark_results(
                bench_cfg, time_src, bench_cfg_sample_hash, run_cfg
            )
            if self.sample_cache is not None:
                self.sample_cache.close()

            # use the hash of the inputs to look up historical values in the cache
            if run_cfg.over_time:
                bench_cfg.ds = self.load_history_cache(
                    bench_cfg.ds, bench_cfg_hash, run_cfg.clear_history
                )

            self.report_results(bench_cfg, run_cfg.print_xarray, run_cfg.print_pandas)
            self.cache_results(bench_cfg, bench_cfg_hash)

        self.pane = BenchPlotter.plot(bench_cfg, self.pane)
        return bench_cfg

    def check_var_is_a_param(self, variable: param.Parameter, var_type: str):
        """check that a variable is a subclass of param

        Args:
            variable (param.Parameter): the varible to check
            var_type (str): a string representation of the variable type for better error messages

        Raises:
            TypeError: the input variable type is not a param.
        """
        if not isinstance(variable, param.Parameter):
            raise TypeError(
                f"You need to use {var_type}_vars =[{self.worker_input_cfg}.param.your_variable], instead of {var_type}_vars =[{self.worker_input_cfg}.your_variable]"
            )

    def cache_results(self, bench_cfg: BenchCfg, bench_cfg_hash: int) -> None:
        with Cache("cachedir/benchmark_inputs") as c:
            logging.info(f"saving results with key: {bench_cfg_hash}")
            self.bench_cfg_hashes.append(bench_cfg_hash)
            c[bench_cfg_hash] = bench_cfg
            logging.info(f"saving benchmark: {self.bench_name}")
            c[self.bench_name] = self.bench_cfg_hashes

    def calculate_benchmark_results(
        self, bench_cfg, time_src: datetime | str, bench_cfg_sample_hash, bench_run_cfg
    ):
        """A function for generating an n-d xarray from a set of input variables in the BenchCfg

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            time_src (datetime): a representation of the sample time

        Returns:
            bench_cfg (BenchCfg): description of the benchmark parameters
        """
        bench_cfg, func_inputs, dims_name = self.setup_dataset(bench_cfg, time_src)
        constant_inputs = self.define_const_inputs(bench_cfg.const_vars)
        callcount = 1
        bench_cfg.hmap_kdims = dims_name
        for idx_tuple, function_input_vars in func_inputs:
            logging.info(f"{bench_cfg.title}:call {callcount}/{len(func_inputs)}")
            self.call_worker_and_store_results(
                bench_cfg,
                idx_tuple,
                function_input_vars,
                dims_name,
                constant_inputs,
                bench_cfg_sample_hash,
                bench_run_cfg,
            )
            callcount += 1

        for inp in bench_cfg.all_vars:
            self.add_metadata_to_dataset(bench_cfg, inp)
        return bench_cfg

    def plot(self, run_cfg: BenchRunCfg = None) -> None:
        """Launches a webserver with plots of the benchmark results, blocking

        Args:
            run_cfg (BenchRunCfg, optional): Options for the webserve such as the port. Defaults to None.

        """
        if run_cfg is None:
            run_cfg = self.last_run_cfg
        BenchPlotServer().plot_server(self.bench_name, run_cfg, self.pane)

    def load_history_cache(
        self, ds: xr.Dataset, bench_cfg_hash: int, clear_history: bool
    ) -> xr.Dataset:
        """Load historical data from a cache if over_time=true

        Args:
            ds (xr.Dataset): Freshly calcuated data
            bench_cfg_hash (int): Hash of the input variables used to generate the data
            clear_history (bool): Optionally clear the history

        Returns:
            xr.Dataset: historical data as an xr dataset
        """
        with Cache("cachedir/history") as c:
            if clear_history:
                logging.info("clearing history")
            else:
                logging.info(f"checking historical key: {bench_cfg_hash}")
                if bench_cfg_hash in c:
                    logging.info("loading historical data from cache")
                    ds_old = c[bench_cfg_hash]
                    ds = xr.concat([ds_old, ds], "over_time")
                else:
                    logging.info("did not detect any historical data")

            logging.info("saving data to history cache")
            c[bench_cfg_hash] = ds
        return ds

    def setup_dataset(
        self, bench_cfg: BenchCfg, time_src: datetime | str
    ) -> tuple[BenchCfg, List, List]:
        """A function for generating an n-d xarray from a set of input variables in the BenchCfg

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            time_src (datetime | str): a representation of the sample time

        Returns:
            _type_: _description_
        """

        if time_src is None:
            time_src = datetime.now()
        bench_cfg.meta_vars = self.define_extra_vars(bench_cfg, bench_cfg.repeats, time_src)

        bench_cfg.all_vars = bench_cfg.input_vars + bench_cfg.meta_vars

        for i in bench_cfg.all_vars:
            logging.info(i.sampling_str(bench_cfg.debug))

        dims_cfg = DimsCfg(bench_cfg)
        function_inputs = list(
            zip(product(*dims_cfg.dim_ranges_index), product(*dims_cfg.dim_ranges))
        )
        # xarray stores K N-dimensional arrays of data.  Each array is named and in this case we have a nd array for each result variable
        data_vars = {}

        for rv in bench_cfg.result_vars:
            if type(rv) == ResultVar:
                result_data = np.empty(dims_cfg.dims_size)
                result_data.fill(np.nan)
                data_vars[rv.name] = (dims_cfg.dims_name, result_data)
            elif type(rv) == ResultVec:
                for i in range(rv.size):
                    result_data = np.full(dims_cfg.dims_size, np.nan)
                    data_vars[rv.index_name(i)] = (dims_cfg.dims_name, result_data)

        bench_cfg.ds = xr.Dataset(data_vars=data_vars, coords=dims_cfg.coords)
        bench_cfg.ds_dynamic = self.ds_dynamic

        return bench_cfg, function_inputs, dims_cfg.dims_name

    def define_const_inputs(self, const_vars) -> dict:
        constant_inputs = None
        if const_vars is not None:
            const_vars, constant_values = [
                [i for i, j in const_vars],
                [j for i, j in const_vars],
            ]

            constant_names = [i.name for i in const_vars]
            constant_inputs = dict(zip(constant_names, constant_values))
        return constant_inputs

    def define_extra_vars(self, bench_cfg: BenchCfg, repeats: int, time_src) -> list[IntSweep]:
        """Define extra meta vars that are stored in the n-d array but are not passed to the benchmarking function, such as number of repeats and the time the function was called.

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            repeats (int): the number of times to sample the function
            time_src (datetime): a representation of the sample time

        Returns:
            _type_: _description_
        """
        bench_cfg.iv_repeat = IntSweep(
            default=repeats,
            bounds=[1, repeats],
            samples=repeats,
            samples_debug=2 if repeats > 2 else 1,
            units="repeats",
        )
        bench_cfg.iv_repeat.name = "repeat"
        extra_vars = [bench_cfg.iv_repeat]

        if bench_cfg.over_time:
            if type(time_src) == str:
                iv_over_time = TimeEvent(time_src)
            else:
                iv_over_time = TimeSnapshot(time_src)
            iv_over_time.name = "over_time"
            extra_vars.append(iv_over_time)
            bench_cfg.iv_time = [iv_over_time]
        return extra_vars

    def worker_wrapper(self, bench_cfg: BenchCfg, function_input: dict):
        self.worker_fn_call_count += 1
        if not bench_cfg.pass_repeat:
            function_input.pop("repeat")
        if "over_time" in function_input:
            function_input.pop("over_time")
        if "time_event" in function_input:
            function_input.pop("time_event")

        if self.worker_input_cfg is None:  # worker takes kwargs
            return self.worker(**function_input)

        # worker takes a parametrised input object
        input_cfg = self.worker_input_cfg()
        for k, v in function_input.items():
            input_cfg.param.set_param(k, v)

        return self.worker(input_cfg)

    def call_worker_and_store_results(
        self,
        bench_cfg: BenchCfg,
        index_tuple: tuple,
        function_input_vars: List,
        dims_name: List[str],
        constant_inputs: dict,
        bench_cfg_sample_hash: str,
        bench_run_cfg: BenchRunCfg,
    ) -> None:
        """A wrapper around the benchmarking function to set up and store the results of the benchmark function

        Args:
            bench_cfg (BenchCfg): description of the benchmark parameters
            index_tuple (tuple): tuple of n-d array indices
            function_input_vars (List): list of function inputs as dicts
            dims_name (List[str]): names of the n-d dimension
            constant_inputs (dict): input values to keep constant
        """
        self.worker_wrapper_call_count += 1
        function_input = dict(zip(dims_name, function_input_vars))

        fn_inp_with_rep = tuple(function_input.values())

        if constant_inputs is not None:
            function_input |= constant_inputs

        if bench_cfg.print_bench_inputs:
            logging.info("Bench Inputs:")
            for k, v in function_input.items():
                logging.info(f"\t {k}:{v}")

        # store a tuple of the inputs as keys for a holomap
        if bench_cfg.use_sample_cache:
            # the signature is the hash of the inputs to to the function + meta variables such as repeat and time + the hash of the benchmark sweep as a whole (without the repeats hash)
            fn_inputs_sorted = list(SortedDict(function_input).items())
            function_input_signature_pure = hash_sha1((fn_inputs_sorted, bench_cfg.tag))

            function_input_signature_benchmark_context = hash_sha1(
                (function_input_signature_pure, bench_cfg_sample_hash)
            )
            print("inputs", fn_inputs_sorted)
            print("pure", function_input_signature_pure)
            if function_input_signature_benchmark_context in self.sample_cache:
                logging.info(
                    f"Found a previously calculated value in the sample cache with the benchmark: {bench_cfg.title}, hash: {bench_cfg_sample_hash}"
                )
                result = self.sample_cache[function_input_signature_benchmark_context]
                self.worker_cache_call_count += 1
            elif bench_run_cfg.only_hash_tag and function_input_signature_pure in self.sample_cache:
                logging.info(
                    f"A value including the benchmark context was not found: {bench_cfg.title} hash: {bench_cfg_sample_hash}, but was found with tag:{bench_cfg.tag} so loading those values from the cache.  Beware that depending on how you have run the benchmarks, the data in this cache could be invalid"
                )
                result = self.sample_cache[function_input_signature_pure]
                self.worker_cache_call_count += 1
            else:
                logging.info(
                    "Sample cache values Not Found for either pure function inputs or inputs within a benchmark context, calling benchmark function"
                )
                result = self.worker_wrapper(bench_cfg, function_input)
                self.sample_cache.set(
                    function_input_signature_benchmark_context, result, tag=bench_cfg.tag
                )
                self.sample_cache.set(function_input_signature_pure, result, tag=bench_cfg.tag)
        else:
            result = self.worker_wrapper(bench_cfg, function_input)

        # construct a dict for a holomap
        if type(result) == dict:  # todo holomaps with named types
            if "hmap" in result:
                # print(isinstance(result["hmap"], hv.element.Element))
                bench_cfg.hmap[fn_inp_with_rep] = result["hmap"]

        logging.debug(f"input_index {index_tuple}")
        logging.debug(f"input {function_input_vars}")
        logging.debug(f"result {result}")
        for rv in bench_cfg.result_vars:
            if type(result) == dict:
                result_value = result[rv.name]
            else:
                result_value = result.param.values()[rv.name]

            if bench_run_cfg.print_bench_results:
                logging.info(f"{rv.name}: {result_value}")

            if type(rv) == ResultVar:
                set_xarray_multidim(bench_cfg.ds[rv.name], index_tuple, result_value)
            elif type(rv) == ResultVec:
                if isinstance(result_value, (list, np.ndarray)):
                    if len(result_value) == rv.size:
                        for i in range(rv.size):
                            set_xarray_multidim(
                                bench_cfg.ds[rv.index_name(i)], index_tuple, result_value[i]
                            )
            elif type(rv) == ResultList:
                # TODO generalise this for arbirary dimensions, only works for 1 input dim so far.
                dimname = bench_cfg.input_vars[0].name
                input_value = function_input[dimname]

                new_dataarray = xr.DataArray(
                    [result_value.values],
                    name=rv.name,
                    coords={dimname: [input_value], rv.dim_name: result_value.index},
                )

                dataset_key = (bench_cfg.hash_value, rv.name)
                if dataset_key in self.ds_dynamic:
                    self.ds_dynamic[dataset_key] = xr.concat(
                        [self.ds_dynamic[dataset_key], new_dataarray], dimname
                    )
                else:
                    self.ds_dynamic[dataset_key] = new_dataarray
            else:
                raise RuntimeError("Unsupported result type")

    def clear_tag_from_cache(self, tag: str):
        """Clear all samples from the cache that match a tag
        Args:
            tag(str): clear samples with this tag
        """
        if self.sample_cache is None:
            self.sample_cache = Cache("cachedir/sample_cache", tag_index=True)
        logging.info(f"clearing the sample cache for tag: {tag}")
        removed_vals = self.sample_cache.evict(tag)
        logging.info(f"removed: {removed_vals} items from the cache")

    def add_metadata_to_dataset(self, bench_cfg: BenchCfg, input_var: ParametrizedSweep) -> None:
        """Adds variable metadata to the xrarry so that it can be used to automatically plot the dimension units etc.

        Args:
            bench_cfg (BenchCfg):
            input_var (ParametrizedSweep): The varible to extract metadata from
        """

        for rv in bench_cfg.result_vars:
            if type(rv) == ResultVar:
                bench_cfg.ds[rv.name].attrs["units"] = rv.units
                bench_cfg.ds[rv.name].attrs["long_name"] = rv.name
            elif type(rv) == ResultVec:
                for i in range(rv.size):
                    bench_cfg.ds[rv.index_name(i)].attrs["units"] = rv.units
                    bench_cfg.ds[rv.index_name(i)].attrs["long_name"] = rv.name
            else:
                pass  # todo

        dsvar = bench_cfg.ds[input_var.name]
        dsvar.attrs["long_name"] = input_var.name
        if input_var.units is not None:
            dsvar.attrs["units"] = input_var.units
        if input_var.__doc__ is not None:
            dsvar.attrs["description"] = input_var.__doc__

    def report_results(self, bench_cfg: BenchCfg, print_xarray: bool, print_pandas: bool):
        """Optionally display the caculated benchmark data as either as pandas, xarray or plot

        Args:
            bench_cfg (BenchCfg):
            print_xarray (bool):
            print_pandas (bool):
        """
        if print_xarray:
            logging.info(bench_cfg.ds)
        if print_pandas:
            logging.info(bench_cfg.ds.to_dataframe())

    def clear_call_counts(self) -> None:
        """Clear the worker and cache call counts, to help debug and assert caching is happening properly"""
        self.worker_wrapper_call_count = 0
        self.worker_fn_call_count = 0
        self.worker_cache_call_count = 0

    def get_panel(
        self,
        main_plot: bool = True,
    ) -> pn.pane:
        """Get the panel instance where bencher collates results

        Args:
            main_plot (bool, optional): return the last added page. Defaults to True.

        Returns:
            pn.pane: results panel
        """
        if main_plot:
            return self.pane[-1][0]
        return self.pane[-1]

    def append(self, pane: pn.panel) -> None:
        self.get_panel().append(pane)

    def append_tab(self, pane: pn.panel):
        self.pane.append(pane)

    def get_best_params(self, bench_cfg: BenchCfg) -> dict:
        """Get a dictionary of the best found parameters found during the sweep

        Args:
            bench_cfg (BenchCfg): Results

        Returns:
            dict: dictionary of best params
        """
        return bench_cfg.studies[0].best_trials[0].params



import matplotlib.pyplot as plt
import panel as pn
import seaborn as sns
from typing import Optional

from bencher.plotting.plot_filter import PlotFilter, PlotInput, VarRange
from bencher.plotting.plot_types import PlotTypes
from bencher.plotting_functions import wrap_long_time_labels


class Scatter:
    """A class to display the result data as a scatter plot"""

    def scatter2D_sns(self, pl_in: PlotInput) -> Optional[pn.panel]:
        """Given a benchCfg generate a 2D scatter plot from seaborn

        Args:
            pl_in (PlotInput): data to plot

        Returns:
            Optional[pn.pane]: A panel pane with a scatter plot
        """

        if PlotFilter(
            float_range=VarRange(None, None),
            cat_range=VarRange(None, None),
            vector_len=VarRange(2, 2),
        ).matches(pl_in.plt_cnt_cfg):
            bench_cfg = wrap_long_time_labels(pl_in.bench_cfg)
            ds = bench_cfg.ds.drop_vars("repeat")

            df = ds.to_dataframe().reset_index()

            names = pl_in.rv.index_names()

            if bench_cfg.input_vars:
                h = sns.jointplot(df, x=names[0], y=names[1], hue=bench_cfg.input_vars[0].name)
            elif bench_cfg.over_time:
                h = sns.jointplot(df, x=names[0], y=names[1], hue=bench_cfg.iv_time[0].name)

            else:
                h = sns.jointplot(df, x=names[0], y=names[1])

            h.set_axis_labels(f"{names[0]} [{pl_in.rv.units}]", f"{names[1]} [{pl_in.rv.units}]")
            plt.tight_layout()
            return pn.panel(h.fig, name=PlotTypes.scatter2D_sns)
        return None



import unittest
import pytest
from bencher import BenchRunCfg
from bencher.example.example_categorical import example_categorical
from bencher.example.example_floats import example_floats
from bencher.example.example_floats2D import example_floats2D
from bencher.example.example_pareto import example_pareto
from bencher.example.example_simple_cat import example_1D_cat
from bencher.example.example_simple_float import example_1D_float
from bencher.example.example_persistent import example_persistent
from bencher.example.example_float_cat import example_cat_float
from bencher.example.example_time_event import example_time_event
from bencher.example.example_float3D import example_floats3D
from bencher.example.example_float3D_cone import example_cone
from bencher.example.example_custom_sweep import example_custom_sweep
from bencher.example.example_workflow import example_floats2D_workflow, example_floats3D_workflow
from bencher.example.example_vector import example_vector
from bencher.example.example_plot_library import example_plot_library


class TestBenchExamples(unittest.TestCase):
    """The purpose of this test class is to run the example problems to make sure they don't crash.  The bencher logic is tested in the other test files test_bencher.py and test_vars.py"""

    def create_run_cfg(self) -> BenchRunCfg:
        cfg = BenchRunCfg()
        cfg.repeats = 2  # low number of repeats to reduce test time, but also test averaging and variance code
        cfg.debug = True  # reduce size of param sweep so tests are faster
        cfg.clear_cache = True
        return cfg

    def test_example_categorical(self) -> None:
        self.assertIsNotNone(example_categorical(self.create_run_cfg()))

    def test_example_floats(self) -> None:
        self.assertIsNotNone(example_floats(self.create_run_cfg()))

    def test_example_floats2D(self) -> None:
        self.assertIsNotNone(example_floats2D(self.create_run_cfg()))

    def test_example_pareto(self) -> None:
        self.assertIsNotNone(example_pareto(self.create_run_cfg()))

    def test_example_simple_cat(self) -> None:
        self.assertIsNotNone(example_1D_cat(self.create_run_cfg()))

    def test_example_simple_float(self) -> None:
        self.assertIsNotNone(example_1D_float(self.create_run_cfg()))

    def test_example_persistent(self) -> None:
        self.assertIsNotNone(example_persistent(self.create_run_cfg()))

    def test_example_cat_float(self) -> None:
        self.assertIsNotNone(example_cat_float(self.create_run_cfg()))

    def test_example_time_event(self) -> None:
        self.assertIsNotNone(example_time_event(self.create_run_cfg()))

    def test_example_float3D(self) -> None:
        self.assertIsNotNone(example_floats3D(self.create_run_cfg()))

    def test_example_cone(self) -> None:
        self.assertIsNotNone(example_cone(self.create_run_cfg()))

    def test_example_custom_sweep(self) -> None:
        self.assertIsNotNone(example_custom_sweep(self.create_run_cfg()))

    def test_example_floats2D_workflow(self) -> None:
        self.assertIsNotNone(example_floats2D_workflow(self.create_run_cfg()))

    def test_example_floats3D_workflow(self) -> None:
        self.assertIsNotNone(example_floats3D_workflow(self.create_run_cfg()))

    @pytest.mark.skip
    def test_example_vector(self) -> None:
        self.assertIsNotNone(example_vector())

    def test_plot_library(self) -> None:
        self.assertIsNotNone(example_plot_library(self.create_run_cfg()))



import unittest
from hypothesis import given, strategies as st  # pylint: disable=unused-import
import pytest
from bencher.bench_vars import IntSweep


class TestVarSweeps(unittest.TestCase):
    def test_int_sweep_01(self):
        int_sweep = IntSweep(bounds=[0, 1])
        self.assertEqual(int_sweep.default, 0)
        self.assertEqual(int_sweep.values(False), [0, 1])
        self.assertEqual(int_sweep.values(True), [0, 1])

    def test_int_sweep_06(self):
        int_sweep = IntSweep(bounds=[0, 6])
        self.assertEqual(int_sweep.default, 0)
        self.assertEqual(int_sweep.values(False), [0, 1, 2, 3, 4, 5, 6])
        self.assertEqual(int_sweep.values(True), [0, 6])

    def test_int_sweep_06_debug_sampes(self):
        int_sweep = IntSweep(bounds=[0, 6], samples_debug=4)
        self.assertEqual(int_sweep.default, 0)
        self.assertEqual(int_sweep.values(False), [0, 1, 2, 3, 4, 5, 6])
        self.assertEqual(int_sweep.values(True), [0, 2, 4, 6])

    @given(st.integers(min_value=1, max_value=10))
    def test_int_sweep_samples(self, samples):
        int_sweep = IntSweep(bounds=[0, 10], samples=samples)
        self.assertEqual(int_sweep.default, 0)
        self.assertEqual(len(int_sweep.values(False)), samples)

    @given(st.integers(min_value=1, max_value=10))
    def test_int_sweep_debug_samples(self, samples_debug):
        int_sweep = IntSweep(bounds=[0, 10], samples_debug=samples_debug)
        self.assertEqual(int_sweep.default, 0)
        self.assertEqual(len(int_sweep.values(True)), samples_debug)

    @pytest.mark.skip
    @given(st.integers(min_value=1, max_value=10), st.integers(min_value=1, max_value=10))
    def test_int_sweep_samples_all(self, samples, samples_debug):
        int_sweep = IntSweep(bounds=[0, 10], samples=samples)
        self.assertEqual(int_sweep.default, 0)
        self.assertEqual(len(int_sweep.values(False)), samples)
        self.assertEqual(len(int_sweep.values(True)), samples_debug)



# THIS IS NOT A WORKING EXAMPLE YET

# pylint: disable=duplicate-code,unused-argument


import bencher as bch
import math
import random
import numpy as np
import panel as pn
import holoviews as hv


from strenum import StrEnum
from enum import auto


class Function(StrEnum):
    fn_cos = auto()
    fn_sin = auto()


class Waves(bch.ParametrizedSweep):
    phase = bch.FloatSweep(
        default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=2
    )

    freq = bch.FloatSweep(default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=2)

    theta = bch.FloatSweep(
        default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=10
    )

    compute_fn = bch.EnumSweep(Function)

    noisy = bch.BoolSweep(default=True)

    fn_output = bch.ResultVar(units="v", doc="sin of theta with some noise")
    # out_cos = bch.ResultVar(units="v", doc="cos of theta with some noise")

    out_sum = bch.ResultVar(units="v", doc="The sum")

    def calc(
        self, phase=0.0, freq=1.0, theta=0.0, noisy=True, compute_fn: Function = Function.fn_sin
    ) -> dict:
        if noisy:
            noise = 1.0
        else:
            noise = 0.0

        match compute_fn:
            case Function.fn_sin:
                self.fn_output = np.sin(phase + freq * theta)
            case Function.fn_cos:
                self.fn_output = np.cos(phase + freq * theta)

        self.fn_output += random.uniform(0, noise)

        pt = hv.Text(0, 0, f"{phase}\n{freq}\n {theta}")
        pt *= hv.Ellipse(0, 0, 1)
        pt = hv.Points([theta, self.fn_output])

        return self.get_results_values_as_dict(holomap=pt)

    def calc_vec(
        self, phase=0.0, freq=1.0, noisy=True, compute_fn: Function = Function.fn_sin
    ) -> dict:
        if noisy:
            noise = 1.0
        else:
            noise = 0.0

        theta = np.arange(0, 6, 0.1)

        match compute_fn:
            case Function.fn_sin:
                dat = np.sin(phase + freq * theta) + random.uniform(0, noise)
            case Function.fn_cos:
                dat = np.cos(phase + freq * theta) + random.uniform(0, noise)

        self.out_sum = sum(dat)
        hv.Curve((theta, dat), "theta", "voltage")

        pt = hv.Text(0, 0, f"{compute_fn}\n{phase}\n{freq}")
        pt *= hv.Ellipse(0, 0, 1)

        return self.get_results_values_as_dict(holomap=pt)


if __name__ == "__main__":
    # opts.defaults(
    #     # opts.NdLayout(shared_axes=False, shared_datasource=False, width=500),
    #     opts.Overlay(width=500, legend_position="right"),
    # )
    wv = Waves()

    bch_wv = bch.Bench("waves", wv.calc, plot_lib=None)

    res = bch_wv.plot_sweep(
        "phase",
        input_vars=[wv.param.theta, wv.param.freq, wv.param.phase],
        result_vars=[wv.param.fn_output],
        run_cfg=bch.BenchRunCfg(repeats=3),
    )
    bch_wv.append_tab(wv.to_dynamic_map(wv.calc))

    bch_wv.worker = wv.calc_vec
    res = bch_wv.plot_sweep(
        "holo",
        input_vars=[wv.param.freq, wv.param.phase, wv.param.compute_fn],
        result_vars=[wv.param.out_sum],
        run_cfg=bch.BenchRunCfg(repeats=3, auto_plot=False),
    )

    # bch_wv.append(res.to_heatmap())
    # bch_wv.append(res.to_holomap())
    # bch_wv.append_tab(res.to_nd_layout())  # doesn't work on the same page yet.. TODO

    bch_wv.append(res.to_grid())

    bch_wv.append_tab(wv.to_dynamic_map(wv.calc_vec, "theta"))

    # bch_wv.append_tab(wv.to_holomap(wv.calc_vec, "theta"))

    # bch_wv.append(hv.HoloMap(res.hmap, res.hmap_kdims).opts(backend="bokeh"))
    # bch_wv.plots_instance.append(res.to_grid().opts(backend="bokeh"))
    # bch_wv.append(res.to_holomap().layout().)

    # bch_wv.append(res.to_holomap())
    # bch_wv.append(res.to_grid())

    # bch_wv.append(ndlay.grid("freq"))
    bch_wv.plot()

    # bch_wv.append(res.to_curve().layout())

    # res = bch_wv.plot_sweep(
    #     "phase",
    #     input_vars=[wv.param.theta, wv.param.freq, wv.param.phase, wv.param.noisy],
    #     # input_vars=[wv.param.theta, wv.param.freq, wv.param.phase],
    #     result_vars=[wv.param.fn_output],
    #     run_cfg=bch.BenchRunCfg(repeats=5),
    # )

    # bch_wv.plots_instance.append(res.to_curve())
    # bch_wv.plots_instance.append(res.to_curve().overlay("phase"))
    # bch_wv.plots_instance.append(res.to_curve().overlay("freq"))
    # bch_wv.plots_instance.append(res.to_curve().overlay("phase").layout())
    # bch_wv.plots_instance.append(res.to_curve().overlay("freq").layout())
    # bch_wv.plots_instance.append(res.to_curve().layout("phase"))
    # bch_wv.append(res.to_curve().overlay("phase"))
    gspec = pn.GridSpec()

    c1 = res.to_curve()  #
    c1 += res.to_curve().overlay("phase")
    # c1 += res.to_curve().overlay("freq")
    bch_wv.append(c1)
    # gspec[0, 0] = res.to_curve()
    # gspec[0, 1] = res.to_curve().overlay("phase")
    # gspec[1, 0] = res.to_curve().overlay("freq")

    bch_wv.append(gspec)

    # bch_wv.append(res.to_curve().layout("phase"))

    # bch_wv.append(res.to_curve().layout("freq"))
    # bch_wv.append(res.to_curve().layout())

    bch_wv.plot()


# todo  https://discourse.holoviz.org/t/pointdraw-as-parameterized-class/3539



import bencher as bch


class InputCfg(bch.ParametrizedSweep):
    """An example of a datatype with an integer and float parameter"""

    x = bch.FloatSweep(
        sample_values=[2, 3, 4, 7, 8, 9],
        doc="An example of a user defines set of sweep values",
    )
    y = bch.IntSweep(
        sample_values=[1, 2, 6], doc="An example of a user defines set of sweep values"
    )


class Result(bch.ParametrizedSweep):
    result = bch.ResultVar("ul", doc="Square of x")


def benchmark_fn(cfg: InputCfg) -> Result:
    output = Result()
    output.result = cfg.x * cfg.x - cfg.y * 3
    return output


def example_custom_sweep(run_cfg: bch.BenchRunCfg) -> bch.Bench:
    """This example shows how to define a custom set of value to sample from intead of a uniform sweep

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """

    bencher = bch.Bench("benchmarking_example_custom_sweep", benchmark_fn, InputCfg)

    # here we sample the input variable theta and plot the value of output1. The (noisy) function is sampled 20 times so you can see the distribution

    bencher.plot_sweep(
        title="Example User Defined Sweep 1D",
        # input_vars=[InputCfg.param.x],
        input_vars=[InputCfg.param.x],
        result_vars=[Result.param.result],
        description=example_custom_sweep.__doc__,
        run_cfg=run_cfg,
    )

    bencher.plot_sweep(
        title="Example User Defined Sweep 2D",
        # input_vars=[InputCfg.param.x],
        input_vars=[InputCfg.param.x, InputCfg.param.y],
        result_vars=[Result.param.result],
        description=example_custom_sweep.__doc__,
        run_cfg=run_cfg,
    )
    return bencher


if __name__ == "__main__":
    ex_run_cfg = bch.BenchRunCfg()
    # ex_run_cfg.debug = True
    example_custom_sweep(ex_run_cfg).plot()



import numpy as np

import bencher as bch


class VolumeSample(bch.ParametrizedSweep):
    """A class to represent a 3D point in space."""

    pos_samples = 1

    x = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], samples=pos_samples, doc="x coordinate of the sample volume"
    )
    y = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], samples=pos_samples, doc="y coordinate of the sample volume"
    )
    z = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], samples=pos_samples, doc="z coordinate of the sample volume"
    )

    vec_samples = 5

    vec_x = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], samples=vec_samples, doc="x coordinate of the sample volume"
    )
    vec_y = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], samples=vec_samples, doc="y coordinate of the sample volume"
    )
    vec_z = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], samples=vec_samples, doc="z coordinate of the sample volume"
    )


class VolumeResult(bch.ParametrizedSweep):
    """A class to represent the properties of a volume sample."""

    vec_dir = bch.ResultVec(3, "vec", doc="A vector field with an interesting shape")


def bench_fn(point: VolumeSample, normalise=False) -> VolumeResult:
    """This function takes a 3D point as input and returns distance of that point to the origin.

    Args:
        point (VolumeSample): Sample point

    Returns:
        VolumeResult: Value at that point
    """
    output = VolumeResult()

    vec = np.array([point.vec_x, point.vec_y, point.vec_z])

    if normalise:
        norm = np.linalg.norm(vec)
        if norm > 0:
            vec /= norm

    output.vec_dir = list(vec)
    return output


def example_cone(run_cfg: bch.BenchRunCfg) -> bch.Bench:
    """Example of how to perform a 3D floating point parameter sweep

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """
    bench = bch.Bench("Bencher_Example_Cone", bench_fn, VolumeSample)

    bench.plot_sweep(
        input_vars=[
            VolumeSample.param.x,
            VolumeSample.param.y,
            VolumeSample.param.z,
            VolumeSample.param.vec_x,
            VolumeSample.param.vec_y,
            VolumeSample.param.vec_z,
        ],
        result_vars=[
            VolumeResult.param.vec_dir,
        ],
        title="Float 3D cone Example",
        description="""This example shows how to sample 3 floating point variables and plot a vector field representation of the results.  The benchmark function returns the distance to the origin""",
        run_cfg=run_cfg,
    )

    return bench


if __name__ == "__main__":
    ex_run_cfg = bch.BenchRunCfg()
    ex_run_cfg.use_cache = True
    example_cone(ex_run_cfg).plot()



# you need this import to be able to reference a class from a static method in that class
from __future__ import annotations

import math
from pickle import PicklingError, UnpicklingError
from typing import Type

# """This file has some examples for how to perform basic benchmarking parameter sweeps"""
from bencher import Bench, BenchRunCfg, BoolSweep, FloatSweep, ParametrizedSweep
from bencher.example.benchmark_data import ExampleBenchCfgOut


class ExpensiveNonPickleableResource:
    """This represents a global resource that is expensive to setup and teardown and is not serialisable e.g. (A simulator,a connection to a piece of real world hardware etc.)"""

    def __init__(self) -> None:
        self.id = None

    def __getstate__(self) -> Type[PicklingError]:
        """Make sure this is not pickleable"""
        raise PicklingError

    def __setstate__(self, from_state) -> Type[UnpicklingError]:
        """Make sure this is not pickleable"""
        raise UnpicklingError

    def setup(self) -> None:
        print("setup expensive resource")
        self.id = 1

    def shutdown(self) -> None:
        print("shutdown expensive resource once")
        self.id = None


# set up as a global variable as it is truely global (i.e real world hardware).  Great care must be taken that the state of this global variable does not corrupt the results of bencher. All meaningful state should be put into an input config so that it can be tracked properly across parameter sweeps


class InputCfg(ParametrizedSweep):
    """A class that has 1 input parameter, but the benchmark function depends on a global resource"""

    theta = FloatSweep(
        default=1,
        bounds=[0, 10],
        doc="The standard deviation of the noise",
        units="v",
    )

    modify_global = BoolSweep(doc="Use global state as part of the function callback")


class GlobalResource:
    def __init__(self):
        self.expensive_res = ExpensiveNonPickleableResource()
        self.expensive_res.setup()

    def shutdown(self):
        self.expensive_res.shutdown()

    def benchmark_function(self, input_cfg: InputCfg) -> ExampleBenchCfgOut:
        output = ExampleBenchCfgOut()
        output.out_sin = math.sin(input_cfg.theta) + self.expensive_res.id
        if input_cfg.modify_global:
            self.expensive_res.id += 1
        return output


def example_persistent(run_cfg: BenchRunCfg) -> Bench:
    """This example shows how maintain a peristent object across runs. This could be useful for resources that are expensive to start up and shut down that can be shared across different function calls during benchmarking. This resource may not be picklable.

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """

    # set up the expensive resource once
    res = GlobalResource()

    bencher = Bench("benchmarking_example_float1D", res.benchmark_function, InputCfg)

    bencher.plot_sweep(
        title="Example 1D Float",
        input_vars=[InputCfg.param.theta, InputCfg.param.modify_global],
        result_vars=[ExampleBenchCfgOut.param.out_sin],
        description=example_persistent.__doc__,
        run_cfg=run_cfg,
    )
    # shutdown the resource at the end
    res.shutdown()

    return bencher


if __name__ == "__main__":
    ex_run_cfg = BenchRunCfg()

    example_persistent(ex_run_cfg).plot()



__version__ = "0.0.0"

from .bencher import Bench, BenchCfg, BenchRunCfg
from .plt_cfg import BenchPlotter, PltCfgBase
from .example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function
from .bench_plot_server import BenchPlotServer
from .bench_vars import (
    IntSweep,
    FloatSweep,
    StringSweep,
    TimeSnapshot,
    EnumSweep,
    BoolSweep,
    ParametrizedSweep,
    ResultVar,
    ResultVec,
    ResultList,
    ResultSeries,
    ResultHmap,
    OptDir,
    hash_sha1,
)
from .plotting.plot_library import PlotLibrary, PlotTypes
from .plotting.plot_filter import PlotInput, VarRange, PlotFilter
from .utils import (
    update_params_from_kwargs,
    get_inputs_only,
    get_results_only,
    get_results_values_as_dict,
    get_input_and_results,
)



# pylint: skip-file
import hashlib
import re
from datetime import datetime
from enum import Enum, auto
from typing import List, Tuple

import numpy as np
import param
from pandas import Timestamp
from param import Boolean, Integer, Number, Parameterized, Selector
from strenum import StrEnum
import holoviews as hv
from collections import namedtuple


def hash_sha1(var: any) -> str:
    """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
    return hashlib.sha1(str(var).encode("ASCII")).hexdigest()


def capitalise_words(message: str):
    """Given a string of lowercase words, capitalise them

    Args:
        message (str): lower case string

    Returns:
        _type_: capitalised string
    """
    capitalized_message = " ".join([word.capitalize() for word in message.split(" ")])
    return capitalized_message


def un_camel(camel: str) -> str:
    """Given a snake_case string return a CamelCase string

    Args:
        camel (str): camelcase string

    Returns:
        str: uncamelcased string
    """

    return capitalise_words(re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", camel.replace("_", " ")))


def param_hash(param_type: Parameterized, hash_value: bool = True, hash_meta: bool = False) -> int:
    """A custom hash function for parametrised types with options for hashing the value of the type and hashing metadata

    Args:
        param_type (Parameterized): A parameter
        hash_value (bool, optional): use the value as part of the hash. Defaults to True.
        hash_meta (bool, optional): use metadata as part of the hash. Defaults to False.

    Returns:
        int: a hash
    """

    curhash = 0
    if hash_value:
        for k, v in param_type.param.values().items():
            if k != "name":
                curhash = hash_sha1((curhash, hash_sha1(v)))

    if hash_meta:
        for k, v in param_type.param.params().items():
            if k != "name":
                print(f"key:{k}, hash:{hash_sha1(k)}")
                print(f"value:{v}, hash:{hash_sha1(v)}")
                curhash = hash_sha1((curhash, hash_sha1(k), hash_sha1(v)))
    return curhash


def make_namedtuple(class_name: str, **fields) -> namedtuple:
    """Convenience method for making a named tuple

    Args:
        class_name (str): name of the named tuple

    Returns:
        namedtuple: a named tuple with the fields as values
    """
    return namedtuple(class_name, fields)(*fields.values())


def as_dim(self, compute_values=False) -> hv.Dimension:
    """Takes a sweep variable and turns it into a holoview dimension

    Returns:
        hv.Dimension:
    """
    if hasattr(self, "bounds"):
        if compute_values:
            return hv.Dimension(
                (self.name, self.name),
                range=tuple(self.bounds),
                unit=self.units,
                values=self.values(False),
            )

        return hv.Dimension(
            (self.name, self.name),
            range=tuple(self.bounds),
            unit=self.units,
            default=self.default,
        )
    return hv.Dimension(
        (self.name, self.name),
        unit=self.units,
        values=self.values(False),
        default=self.default,
    )


class ParametrizedSweep(Parameterized):
    """Parent class for all Sweep types that need a custom hash"""

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return param_hash(self, True, False)

    def update_params_from_kwargs(self, **kwargs) -> None:
        """Given a dictionary of kwargs, set the parameters of the passed class 'self' to the values in the dictionary."""
        used_params = {}
        for key in self.param.params().keys():
            if key in kwargs:
                if key != "name":
                    used_params[key] = kwargs[key]

        self.param.update(**used_params)

    def get_input_and_results(self, include_name: bool = False) -> Tuple[dict, dict]:
        """Get dictionaries of input parameters and result parameters

        Args:
            self: A parametrised class
            include_name (bool): Include the name parameter that all parametrised classes have. Default False

        Returns:
            Tuple[dict, dict]: a tuple containing the inputs and result parameters as dictionaries
        """
        inputs = {}
        results = {}
        for k, v in self.param.params().items():
            if isinstance(v, (ResultList, ResultSeries, ResultVar, ResultVec)):
                results[k] = v
            else:
                inputs[k] = v

        if not include_name:
            inputs.pop("name")
        return make_namedtuple("inputresult", inputs=inputs, results=results)

    def get_results_values_as_dict(self, holomap=None) -> dict:
        """Get a dictionary of result variables with the name and the current value"""
        values = self.param.values()
        output = {key: values[key] for key in self.get_input_and_results().results}
        if holomap is not None:
            output |= {"hmap": holomap}
        return output

    def get_inputs_only(self) -> List[param.Parameter]:
        """Return a list of input parameters

        Returns:
            List[param.Parameter]: A list of input parameters
        """
        return list(self.get_input_and_results().inputs.values())

    def get_inputs_as_dims(
        self, compute_values=False, remove_dims: str | List[str] = None
    ) -> List[hv.Dimension]:
        inputs = self.get_inputs_only()

        if remove_dims is not None:
            if type(remove_dims) == str:
                remove_dims = [remove_dims]
            filtered_inputs = [i for i in inputs if i.name not in remove_dims]
            inputs = filtered_inputs

        return [iv.as_dim(compute_values) for iv in inputs]

    def to_dynamic_map(
        self,
        callback,
        remove_dims: str | List[str] = None,
    ) -> hv.DynamicMap:
        def callback_wrapper(**kwargs):
            return callback(**kwargs)["hmap"]

        return hv.DynamicMap(
            callback=callback_wrapper,
            kdims=self.get_inputs_as_dims(compute_values=False, remove_dims=remove_dims),
        ).opts(shared_axes=False)

    def to_holomap(self, callback, remove_dims: str | List[str] = None) -> hv.DynamicMap:
        return hv.HoloMap(
            hv.DynamicMap(
                callback=callback,
                kdims=self.get_inputs_as_dims(compute_values=True, remove_dims=remove_dims),
            )
        )
        # return hv.HoloMap(self.to_dynamic_map(callback=callback, remove_dims=remove_dims))
        # return hv.DynamicMap(
        #     kdims=self.get_inputs_as_dims(compute_values=True, remove_dims=remove_dims)
        # )


# slots that are shared across all Sweep classes
# param does not work with multiple inheritance so define here
shared_slots = ["units", "samples", "samples_debug"]


def sweep_hash(parameter: Parameterized) -> int:
    """Generate a hash for a sweep variable

    Returns:
        int: hash
    """
    curhash = 0
    for v in parameter.values():
        print(f"value:{v}, hash:{hash_sha1(v)}")
        curhash = hash_sha1((curhash, hash_sha1(v)))
    return curhash


def hash_extra_vars(parameter: Parameterized) -> int:
    """hash extra meta vars in the parameter

    Args:
        parameter (Parameterized): a parameter

    Returns:
        int: hash
    """
    return hash_sha1((parameter.units, parameter.samples, parameter.samples_debug))


def describe_variable(v: Parameterized, debug: bool, include_samples: bool) -> List[str]:
    """Generate a string description of a variable

    Args:
        v (param.Parameterized): parameter to describe
        debug (bool): Generate a reduced number of samples from the variable
        include_samples (bool): Include a description of the samples

    Returns:
        str: String description of the variable
    """
    indent = "    "
    sampling_str = []
    sampling_str.append(f"{v.name}:")
    if include_samples:
        sampling_str.append(f"{indent}{v.sampling_str(debug)}")
    sampling_str.append(f"{indent}units: [{v.units}]")
    if v.doc is not None:
        sampling_str.append(f"{indent}docs: {v.doc}")
    for i in range(len(sampling_str)):
        sampling_str[i] = f"{indent}{sampling_str[i]}"
    return sampling_str


class BoolSweep(Boolean):
    """A class to reprsent a parameter sweep of bools"""

    __slots__ = shared_slots

    def __init__(self, units: str = "ul", samples: int = None, samples_debug: int = 2, **params):
        Boolean.__init__(self, **params)
        self.units = units
        if samples is None:
            self.samples = 2
        self.samples_debug = samples_debug

    def values(self, debug=False) -> List[bool]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        print(self.sampling_str(debug))
        return [True, False]

    def sampling_str(self, debug: bool) -> str:
        """Generate a string representation of the sampling procedure"""
        return f"sampling {self.name} from: [True,False]"

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_extra_vars(self)

    def as_dim(self, compute_values=False) -> hv.Dimension:
        return as_dim(self, compute_values=compute_values)


class TimeBase(Selector):
    """A class to capture a time snapshot of benchmark values.  Time is reprented as a continous value i.e a datetime which is converted into a np.datetime64.  To represent time as a discrete value use the TimeEvent class. The distinction is because holoview and plotly code makes different assumptions about discrete vs continous variables"""

    __slots__ = shared_slots

    def values(self, debug=False) -> List[str]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        print(self.sampling_str(debug))
        return self.objects

    def sampling_str(self, debug: bool) -> str:
        """Generate a string representation of the time sample

        Args:
            debug (bool): If true then self.samples_debug is used
        """
        return f"sampling from [The Past to {self.objects[0]}]"

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_extra_vars(self)


class TimeSnapshot(TimeBase):
    """A class to capture a time snapshot of benchmark values.  Time is reprented as a continous value i.e a datetime which is converted into a np.datetime64.  To represent time as a discrete value use the TimeEvent class. The distinction is because holoview and plotly code makes different assumptions about discrete vs continous variables"""

    __slots__ = shared_slots

    def __init__(
        self,
        datetime_src: datetime | str,
        units: str = "time",
        samples: int = None,
        samples_debug: int = 2,
        **params,
    ):
        if type(datetime_src) == str:
            Selector.__init__(self, [datetime_src], instantiate=True, **params)
        else:
            Selector.__init__(
                self,
                [Timestamp(datetime_src)],
                instantiate=True,
                **params,
            )
        self.units = units
        if samples is None:
            self.samples = len(self.objects)
        else:
            self.samples = samples
        self.samples_debug = min(self.samples, samples_debug)


class TimeEvent(TimeBase):
    """A class to represent a discrete event in time where the data was captured i.e a series of pull requests.  Here time is discrete and can't be interpolated, to represent time as a continous value use the TimeSnapshot class.  The distinction is because holoview and plotly code makes different assumptions about discrete vs continous variables"""

    __slots__ = shared_slots

    def __init__(
        self,
        time_event: str,
        units: str = "event",
        samples: int = None,
        samples_debug: int = 2,
        **params,
    ):
        Selector.__init__(
            self,
            [time_event],
            instantiate=True,
            **params,
        )
        self.units = units
        if samples is None:
            self.samples = len(self.objects)
        else:
            self.samples = samples
        self.samples_debug = min(self.samples, samples_debug)


class StringSweep(Selector):
    """A class to reprsent a parameter sweep of strings"""

    __slots__ = shared_slots

    def __init__(
        self,
        string_list: List[str],
        units: str = "ul",
        samples: int = None,
        samples_debug: int = 2,
        **params,
    ):
        Selector.__init__(self, string_list, instantiate=True, **params)
        self.units = units
        if samples is None:
            self.samples = len(self.objects)
        else:
            self.samples = samples
        self.samples_debug = min(self.samples, samples_debug)

    def values(self, debug=False) -> List[str]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        print(self.sampling_str(debug))
        return self.objects

    def sampling_str(self, debug: bool) -> str:
        """Generate a string representation of the of the sampling procedure

        Args:
            debug (bool): If true then self.samples_debug is used
        """
        object_str = ",".join([i for i in self.objects])
        return f"sampling {self.name} from: [{object_str}]"

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_extra_vars(self)


class EnumSweep(Selector):
    """A class to reprsent a parameter sweep of enums"""

    __slots__ = shared_slots

    def __init__(
        self, enum_type: Enum | List[Enum], units="ul", samples=None, samples_debug=2, **params
    ):
        # The enum can either be an Enum type or a list of enums
        list_of_enums = type(enum_type) is list
        if list_of_enums:
            selector_list = enum_type  # already a list of enums
        else:
            # create a list of enums from the enum type definition
            selector_list = [e for e in enum_type]
        Selector.__init__(self, selector_list, **params)
        if not list_of_enums:  # Grab the docs from the enum type def
            self.doc = enum_type.__doc__
        self.units = units
        if samples is None:
            self.samples = len(self.objects)
        else:
            self.samples = samples
        self.samples_debug = min(self.samples, samples_debug)

    def values(self, debug=False) -> List[Enum]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        print(self.sampling_str(debug))
        return self.objects

    def sampling_str(self, debug: bool) -> str:
        """Generate a string representation of the of the sampling procedure

        Args:
            debug (bool): If true then self.samples_debug is used
        """
        object_str = ",".join([i for i in self.objects])
        return f"sampling {self.name} from: [{object_str}]"

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_extra_vars(self)

    def as_dim(self, compute_values=False) -> hv.Dimension:
        return as_dim(self, compute_values=compute_values)


def int_float_sampling_str(name, samples) -> str:
    """Generate a string representation of the of the sampling procedure

    Args:
        debug (bool): If true then self.samples_debug is used
    """

    return f"sampling {name} from {samples} in {len(samples)} samples"


class IntSweep(Integer):
    """A class to reprsent a parameter sweep of ints"""

    __slots__ = shared_slots + ["sample_values"]

    def __init__(self, units="ul", samples=None, samples_debug=2, sample_values=None, **params):
        Integer.__init__(self, **params)
        self.units = units
        self.samples_debug = samples_debug

        if sample_values is None:
            if samples is None:
                if self.bounds is None:
                    raise RuntimeError("You must define bounds for integer types")
                else:
                    self.samples = 1 + self.bounds[1] - self.bounds[0]
            else:
                self.samples = samples
            self.sample_values = None
        else:
            self.sample_values = sample_values
            self.samples = len(self.sample_values)
            if "default" not in params:
                self.default = sample_values[0]

    def values(self, debug=False) -> List[int]:
        """return all the values for a parameter sweep.  If debug is true return the  list"""
        samps = self.samples_debug if debug else self.samples

        if self.sample_values is None:
            return [
                int(i)
                for i in np.linspace(
                    self.bounds[0], self.bounds[1], samps, endpoint=True, dtype=int
                )
            ]
        else:
            if debug:
                indices = [
                    int(i)
                    for i in np.linspace(
                        0, len(self.sample_values) - 1, self.samples_debug, dtype=int
                    )
                ]
                return [self.sample_values[i] for i in indices]
            else:
                return self.sample_values

    def sampling_str(self, debug: bool) -> str:
        """Generate a string representation of the of the sampling procedure

        Args:
        debug (bool): If true then self.samples_debug is used
        """
        return int_float_sampling_str(self.name, self.values(debug))

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_extra_vars(self)

    def as_dim(self, compute_values=False) -> hv.Dimension:
        return as_dim(self, compute_values=compute_values)

    ###THESE ARE COPIES OF INTEGER VALIDATION BUT ALSO ALLOW NUMPY INT TYPES
    def _validate_value(self, val, allow_None):
        if callable(val):
            return

        if allow_None and val is None:
            return

        if not isinstance(val, (int, np.integer)):
            raise ValueError(
                "Integer parameter %r must be an integer, " "not type %r." % (self.name, type(val))
            )

    ###THESE ARE COPIES OF INTEGER VALIDATION BUT ALSO ALLOW NUMPY INT TYPES
    def _validate_step(self, val, step):
        if step is not None and not isinstance(step, (int, np.integer)):
            raise ValueError(
                "Step can only be None or an " "integer value, not type %r" % type(step)
            )


class FloatSweep(Number):
    """A class to represent a parameter sweep of floats"""

    __slots__ = shared_slots + ["sample_values"]

    def __init__(self, units="ul", samples=10, samples_debug=2, sample_values=None, **params):
        Number.__init__(self, **params)
        self.units = units
        self.samples_debug = samples_debug
        if sample_values is None:
            self.samples = samples
            self.sample_values = None
        else:
            self.sample_values = sample_values
            self.samples = len(self.sample_values)
            if "default" not in params:
                self.default = sample_values[0]

    def values(self, debug=False) -> List[float]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        samps = self.samples_debug if debug else self.samples
        if self.sample_values is None:
            return np.linspace(self.bounds[0], self.bounds[1], samps)
        else:
            if debug:
                indices = [
                    int(i)
                    for i in np.linspace(
                        0, len(self.sample_values) - 1, self.samples_debug, dtype=int
                    )
                ]
                return [self.sample_values[i] for i in indices]
            else:
                return self.sample_values

    def sampling_str(self, debug: bool) -> str:
        """Generate a string representation of the of the sampling procedure

        Args:
            debug (bool): If true then self.samples_debug is used
        """
        return int_float_sampling_str(self.name, self.values(debug))

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_extra_vars(self)

    def as_dim(self, compute_values=False) -> hv.Dimension:
        return as_dim(self, compute_values=compute_values)


class OptDir(StrEnum):
    minimize = auto()
    maximize = auto()
    none = auto()  # If none this var will not appear in pareto plots


class ResultVar(Number):
    """A class to represent result variables and the desired optimisation direction"""

    __slots__ = ["units", "direction"]

    def __init__(self, units="ul", direction: OptDir = OptDir.minimize, **params):
        Number.__init__(self, **params)
        self.units = units
        self.default = 0  # json is terrible and does not support nan values
        self.direction = direction

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_sha1((self.units, self.direction))


class ResultHmap(param.Parameter):
    """A class to represent a holomap return type"""

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_sha1(self)


class ResultVec(param.List):
    """A class to represent fixed size vector result variable"""

    __slots__ = ["units", "direction", "size"]

    def __init__(self, size, units="ul", direction: OptDir = OptDir.minimize, **params):
        param.List.__init__(self, **params)
        self.units = units
        self.default = 0  # json is terrible and does not support nan values
        self.direction = direction
        self.size = size

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_sha1((self.units, self.direction))

    def index_name(self, idx: int) -> str:
        """given the index of the vector, return the column name that

        Args:
            idx (int): index of the result vector

        Returns:
            str: column name of the vector for the xarray dataset
        """

        mapping = ["x", "y", "z"]
        if idx < 3:
            index = mapping[idx]
        else:
            index = idx
        return f"{self.name}_{index}"

    def index_names(self) -> List[str]:
        """Returns a list of all the xarray column names for the result vector

        Returns:
            list[str]: column names
        """
        return [self.index_name(i) for i in range(self.size)]


class ResultSeries:
    """A class to represent a vector of results, it also includes an index similar to pandas.series"""

    def __init__(self, values=None, index=None) -> None:
        self.values = []
        self.index = []
        self.set_index_and_values(values, index)

    def append(self, value: float | int, index: float | int | str = None) -> None:
        """Add a value and index to the result series

        Args:
            value (float | int): result value of the series
            index (float | int | str, optional): index value of series, the same as a pandas.series index  If no value is passed an integer index is automatically created. Defaults to None.
        """

        if index is None:
            self.index.append(len(self.values))
        else:
            self.index.append(index)
        self.values.append(value)

    def set_index_and_values(
        self, values: List[float | int], index: List[float | int | str] = None
    ) -> None:
        """Add values and indices to the result series

        Args:
            value (List[float | int]): result value of the series
            index (List[float | int | str], optional): index value of series, the same as a pandas.series index  If no value is passed an integer index is automatically created. Defaults to None.
        """
        if values is not None:
            if index is None:
                self.index = list(range(len(values)))
            else:
                self.index = index
            self.values = values


class ResultList(param.Parameter):
    """A class to unknown size vector result variable"""

    __slots__ = ["units", "dim_name", "dim_units", "indices"]

    def __init__(
        self,
        index_name: str,
        index_units: str,
        default=ResultSeries(),
        units="ul",
        indices: List[float] = None,
        instantiate=True,
        **params,
    ):
        param.Parameter.__init__(self, default=default, instantiate=instantiate, **params)
        self.units = units
        self.dim_name = index_name
        self.dim_units = index_units
        self.indices = indices

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_sha1((self.units, self.dim_name, self.dim_units))



"""Example on how to sweep over function with vector outputs"""

from math import cos, sin

import numpy as np

import bencher as bch


class OffsetCfg(bch.ParametrizedSweep):
    """A class for describing which parameters to sweep"""

    dc_offset = bch.FloatSweep(
        default=0,
        bounds=[0.0, 2.0],
        samples=4,
        units="v",
        doc="DC offset to add to the result of the signal",
    )
    phase_offset = bch.FloatSweep(
        default=0,
        bounds=[0.0, 3.14],
        samples=4,
        units="rad",
        doc="phase offset that is added to the input before passing to the trig function",
    )


class SweepResult(bch.ParametrizedSweep):
    """A class to describe the vector outputs of the benchmark function"""

    sin_sweep = bch.ResultList(
        index_name="time", index_units="s", units="v", doc="A list of values from a sin function"
    )
    cos_sweep = bch.ResultList(
        index_name="time", index_units="s", units="v", doc="A list of values from a cos function"
    )


def sin_sweep(cfg: OffsetCfg) -> SweepResult:
    """A function that returns vector outputs of the sin and cos functions

    Args:
        cfg (OffsetCfg): Options for phase and dc offset

    Returns:
        SweepResult: vectors with sin and cos results
    """
    res = SweepResult()
    print(type(res.sin_sweep))
    for i in np.arange(0, 6.28, 0.02):
        res.sin_sweep.append(sin(i + cfg.phase_offset) + cfg.dc_offset, i)
        # res.sin_sweep.indices.append(i)
    for i in np.arange(0, 3.28, 0.02):
        res.cos_sweep.append(cos(i + cfg.phase_offset) + cfg.dc_offset, i)
        # res.cos_sweep.indices.append(i)
    return res


def example_vector() -> bch.Bench:
    """Example on how to sweep over function with vector outputs"""
    bencher = bch.Bench("vector output example", sin_sweep, OffsetCfg)

    bencher.plot_sweep(
        title="Sweep DC offset",
        input_vars=[OffsetCfg.param.dc_offset],
        result_vars=[SweepResult.param.sin_sweep, SweepResult.param.cos_sweep],
        description="""This is an example of how to sample function that return a vector of unknown or varying size. In this example it returns the output of the sin and cos function for varying angles and a range of dc offsets""",
        post_description="""The output shows stack of sin and cos functions as the dc offset increases""",
    )

    bencher.plot_sweep(
        title="Sweep phase offset",
        input_vars=[OffsetCfg.param.phase_offset],
        result_vars=[SweepResult.param.sin_sweep, SweepResult.param.cos_sweep],
        description="""This is an example of how to sample function that return a vector of unknown or varying size. In this example it returns the output of the sin and cos function for varying angles and a range of phase offsets""",
        post_description="""The output shows different phases of the sin and cos functions""",
    )

    return bencher


if __name__ == "__main__":
    example_vector().plot()
